/*-------------------------------------------------------------------*
 * Function decim_12k8() and oversamp_12k8()                         *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                         *
 * decim_12k8    : decimation from fs to 12.8kHz.                    *
 * oversamp_12k8 : oversampling from 12.8kHz to fs.                  *
 *                                                                   *
 * fs = 16/22/24/28.8/32/44/48 kHz.                                  *
 *-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "amr_plus_fx.h"

#include "typedef.h"
#include "basic_op.h"
#include "oper_32b.h"
#include "count.h"


#define L_FRAME_MAX   (L_FRAME24k)
#define L_FILT_MAX    (L_FILT24k)

#define L_SUBFR16k   80       /* Subframe size at 16kHz           */


void Oversamp_12k8(
  Word16 sig12k8[],    /* (i) :  signal to oversampling  */
  Word16 sig_fs[],     /* (o) : oversampled signal      */
  Word16 lg,           /* (i) :  length of output        */
  Word16 mem[],        /* (i/o): memory (2*L_FILT)       */
  Word16 band,         /* (i) :  0/2/3=0..6.4k, 1=6.4..10.8k */
  Word16 add          /* (i) :  1 if result is added to  */
                         /*         output vector, 0 if not */
)
{
  Word16 i, L_frame;
  Word16 signal[(2*L_FILT)+L_FRAME_PLUS];
  const Word16 *filter;
  Word16 gain, gain1, den_num;
  Word16 fac_up, fac_down;

  const ncoef = NCOEF_12k8;

  gain = 2;      move16();                                                     
  test();test();test();test();  
  switch (lg) 
  {

    case L_FRAME16kPLUS:
    case L_SUBFR16k:
      fac_up = FAC2_32k;              move16();
      fac_down = FAC1_32k;            move16();
      gain1 = 20480;             move16();
      den_num = 0;            move16();

      test();
      if (band == 0) 
      {
        filter = Filter_32k; move16();
      }
      else
      {
        test();
        if (sub(lg,L_FRAME16kPLUS) == 0) 
        {
          filter = Filter_32k_7k;     move16();
        }
        else
        {
          filter = Filter_32k_hf; move16();
        }
      }    
   
    break;

    case L_FRAME24k:
      fac_up = FAC2_48k;        move16();
      fac_down = FAC1_48k;      move16();
      gain1 = 30720; move16();
      den_num = 0;      move16();
    
      test();
      if (band == 0) 
      {
        filter = Filter_48k; move16();
      }
      else
      {
        filter = Filter_48k_hf; move16();
      }
    break;

    /* this mode is special: it's actually downsampling, thus needs other filter */
    case L_FRAME8k:
      fac_up = FAC2_32k;                    move16();                               
      fac_down = FAC1_32k*2;                move16(); /* note: will be multiplied by 2 again below */ 
      filter = Filter_8k;     move16();                                                         
      gain1 = 20480;                         move16();                               
      den_num = 1;    move16();
    break;

    default:
      printf("wrong frame size in oversamp_12k8! (Unsupported rate)\n");
      exit(0);
  }      

  
  test();
  if (sub(lg,L_FRAME24k) <= 0) 
  {
    fac_down =shl(fac_down,1);
  }
   
  
  test();
  if(sub(lg,L_SUBFR16k) == 0)
  {
    L_frame = L_SUBFR;
  }
  else
  {
    L_frame = L_FRAME_PLUS;                                
  }

  for (i=0; i<(2*ncoef); i++) 
  {
    signal[i] = mem[i];   move16();  
  }

  for (i=0; i<L_frame; i++) 
  {
    signal[i+(2*ncoef)] = mult_r(sig12k8[i],gain1);   move16();        
  }
  
  
  
  Interpol(signal+ncoef, sig_fs, lg, filter, ncoef, 
             fac_up, den_num, fac_down, gain, add);                                
  

  for (i=0; i<(2*ncoef); i++) 
  {
    mem[i] = signal[i+L_frame];   move16();
  }
  return;
}



void Interpol(
  Word16 *signal,         /* (i) : signal to oversampling        */
  Word16 *signal_int,     /* (o) : interpolated signal           */
  Word16 L_frame_int,     /* (i) : Length of interpolated signal */
  const Word16 *filter,   /* (i) : Filter                    Q14 */ 
  Word16 nb_coef,         /* (i) : Number of coeff               */
  Word16 fac_up,          /* (i) : Upsampling factor             */
  Word16 den_num,         /* (i) : Denumerator on numerator      */
  Word16 fac_down,        /* (i) : Downsampling factor           */
  Word16 gain,            /* (i) : Gain to apply                 */
  Word16 Add2            /* (i) ; Add result or not             */
)
{
  Word16 i,ii, frac, frac_step, pos;
  Word16 pos_step_plus_one, pos_step, fac_up_minus_frac_step;
  Word16 *x1, *x2, tmp16;
  Word32 L_sum, Ltmp;
  const Word16 *c1, *c2;

  pos_step = den_num;   move16();
  pos_step_plus_one = add(pos_step , 1);

  frac_step = sub(fac_down, extract_l(L_shr(L_mult(pos_step,fac_up),1)));
  fac_up_minus_frac_step = sub(fac_up,frac_step);

  pos = 0;            move16();
  frac = 0;           move16();

  for(i=0; i<L_frame_int; i++) 
  {
    x1 = signal+pos;                 move16();
    x2 = &x1[1];                 move16(); 
    c1 = &filter[frac];            move16();
    c2 = &filter[sub(fac_up,frac)];    move16();

    L_sum = 0L;                 move32();
    for (ii=0; ii<nb_coef; ii++, c1+=fac_up, c2+=fac_up)
    {
      L_sum = L_mac(L_sum, *x1--, *c1);
      L_sum = L_mac(L_sum, *x2++, *c2);
    }
    L_sum = L_shl(L_sum, add(1,gain));      /* saturation can occur here */
    tmp16 = round(L_sum);

    test();
    if(Add2>0)
    {
      Ltmp = L_mult(tmp16, 32767);
      Ltmp = L_mac(Ltmp, signal_int[i], 16384);      /* Hf synt is always in Q_syn +1*/
      signal_int[i] = round(Ltmp);      move16();
    }
    else
    {
      signal_int[i] = tmp16;                          move16();
    }

    test();
    if (sub(frac,fac_up_minus_frac_step) > 0) 
    {
      pos = add(pos, pos_step_plus_one);
      frac = sub(frac, fac_up_minus_frac_step);
    }
    else 
    {
      pos = add(pos, pos_step);
      frac = add(frac_step, frac);
    }
  }
  return;
}

