#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "typedef.h"
#include "basic_op.h"
#include "count.h"
#include "amr_plus_fx.h"
 
static void ExtractFormat(Word8 *FormatString, Word16 *ReadFlag, Word16 *AppendFlag);

/*____________________________________________________________________________
|
|  FUNCTION NAME: Wave_fopen
|____________________________________________________________________________
*/

FILE *Wave_Fopen(
  Word8 *Filename,
  Word8 *Mode,
  Word16 *NumOfChannels,
  Word32 *SamplingRate,
  Word16 *BitsPerSample,
  Word32 *DataSize
)
{
  FILE *FilePtr, *FilePtrAppend=NULL;
  Word16 ReadFlag;
  Word16 AppendFlag;
  Word8  WaveTag[4];
  Word32  PCM_Size = 0L;
  Word32  FileSize = 0L;      /* Actual size will be replaces in "Wave_fclose" -procedure */
  Word32  FormatSize = 16L;   /* Length of the file format info (in bytes)*/
  Word16 FormatTag = 1;       /* Format Tag (PCM=1) */
  Word16 BytesPerSample;
  Word32  BytesPerSecond;
  Word32  longtmp;
  Word16 shorttmp;
  Word32 sRate;

  ExtractFormat(Mode, &ReadFlag, &AppendFlag);

  test();
  if(ReadFlag == 0)
  {
    test();
    if((FilePtr=fopen(Filename, Mode)) == NULL)
    {
      return FilePtr;
    }
    test();
    if(AppendFlag == 0)
    {
      test();test();
      if( (sub(*BitsPerSample,8) != 0) && (sub(*BitsPerSample,16) != 0) )
      {
        fprintf(stderr, "\n\n");
        fprintf(stderr, "\n ERROR: Only values 16 or 8 allowed for parameter 'BitsPerSample' in function Wave_fopen");
        fprintf(stderr, "\n        [value: %d received]", *BitsPerSample);
        fprintf(stderr, "\n\n");
        exit(EXIT_FAILURE);
      }
      BytesPerSample = (short)(*BitsPerSample /8 * *NumOfChannels);
      BytesPerSecond = *SamplingRate * BytesPerSample;

      BytesPerSample = extract_l(L_shr(L_mult(shr(*BitsPerSample,3), *NumOfChannels),1));
      BytesPerSecond = L_shr(L_mult(extract_l(L_shr(*SamplingRate,4)), BytesPerSample),1-4);

      /* RIFF chunk - 12 bytes */
      fwrite("RIFF", 1, 4, FilePtr);        /* Format*/
      fwrite(&FileSize, 4, 1, FilePtr);     /* Size of the file in bytes (Writen later)*/
      fwrite("WAVE", 1, 4, FilePtr);        /* Format type ID*/

      /* FORMAT chunk - 24 bytes */
      fwrite("fmt ", 1, 4, FilePtr);            /* File format info ID*/
      fwrite(&FormatSize, 4, 1, FilePtr);       /* Length of the FORMAT chunk (in bytes)*/
      fwrite(&FormatTag,  2, 1, FilePtr);       /* Format Tag (PCM = 1)*/
      fwrite(NumOfChannels,  2, 1, FilePtr);    /* Number of channels*/
      sRate = (*SamplingRate);              move16();
      fwrite(&sRate,  4, 1, FilePtr);           /* Sampling rate (Hz)*/
      /*fwrite(SamplingRate,  4, 1, FilePtr);    Sampling rate (Hz)*/
      fwrite(&BytesPerSecond, 4, 1, FilePtr);   /* Number of bytes / second*/
      fwrite(&BytesPerSample,  2, 1, FilePtr);  /* Bytes / sample*/
      fwrite(BitsPerSample,  2, 1, FilePtr);    /* Bits / sample*/

      /* DATA chunk - Data length + 8 bytes */
      fwrite("data",  1, 4, FilePtr);           /* File data ID*/
      fwrite(&PCM_Size,  4, 1, FilePtr);        /* Length of PCM data in bytes*/
    }
    return FilePtr;
  }
  else
  {
    test();
    if((FilePtr=fopen(Filename, Mode)) == NULL)
    {
      return FilePtr;
    }
    test();
    if(AppendFlag>0)
    {
      FilePtrAppend = FilePtr;    move16();
    }
    fseek(FilePtr, 0, SEEK_SET);

    /* RIFF chunk - 12 bytes */
    fread(&WaveTag, 4, 1, FilePtr);
    test();
    if(strncmp("RIFF", WaveTag, 4) != 0)
    {
      return NULL;
    }
    fread(&longtmp, 4, 1, FilePtr);
    fread(&WaveTag, 4, 1, FilePtr);
    test();
    if(strncmp("WAVE", WaveTag, 4) != 0)
    {
      return NULL;
    }

    /* FORMAT chunk - 24 bytes */
    fread(&WaveTag, 4, 1, FilePtr);
    test();
    if(strncmp("fmt ", WaveTag, 4) != 0)
    {
      return NULL;
    }

    fread(&longtmp, 4, 1, FilePtr);
    fread(&shorttmp, 2, 1, FilePtr);
    fread(NumOfChannels,  2, 1, FilePtr);
    fread(SamplingRate,  4, 1, FilePtr);
    fread(&longtmp, 4, 1, FilePtr);
    fread(&BytesPerSample,  2, 1, FilePtr);
    fread(BitsPerSample,  2, 1, FilePtr);

    /* DATA chunk - Data length + 8 bytes */
    fread(&WaveTag, 4, 1, FilePtr);
    test();
    if(strncmp("data", WaveTag, 4) != 0)
    {
      return NULL;
    }
    fread(DataSize,  4, 1, FilePtr);

    *DataSize = *DataSize / BytesPerSample;
    test();  
    if(AppendFlag)
    {
      return FilePtrAppend;
    }
    return FilePtr;
  }
}



/*____________________________________________________________________________
|
|  FUNCTION NAME: Wave_fclose
|____________________________________________________________________________
*/
void Wave_Fclose(
  FILE *FilePtr,
  Word16 BitsPerSample
)
{
  Word32 DataLen;
  Word32 FileLen;
  Word32 StartPos;
  Word32 EndPos;

  test();test();
  if( (sub(BitsPerSample,8) != 0) &&  (sub(BitsPerSample,16) != 0) )
  {
    fprintf(stderr, "\n\n");
    fprintf(stderr, "\n ERROR: Only values 16 or 8 allowed for parameter 'BitsPerSample' in function Wave_fclose");
    fprintf(stderr, "\n\n");
    exit(EXIT_FAILURE);
  }


  EndPos = ftell(FilePtr);
  fseek(FilePtr, 0, SEEK_SET);
  StartPos = ftell(FilePtr);

  FileLen = ((EndPos-StartPos) / (BitsPerSample/16)) - 8;
  DataLen = L_sub(FileLen, 44-8);

  fseek(FilePtr, 4, SEEK_SET);
  fwrite(&FileLen,  1, 4, FilePtr);/* Write length of the wave file*/

  fseek(FilePtr, 40, SEEK_SET);
  fwrite(&DataLen,  1, 4, FilePtr);/* Write longth of the data section*/

  fclose(FilePtr);
}



/*____________________________________________________________________________
|
|  FUNCTION NAME: ExtractFormat
|____________________________________________________________________________
*/
static void ExtractFormat(
  Word8 *FormatString,
  Word16 *ReadFlag,
  Word16 *AppendFlag
)
{
  if (FormatString != NULL)
  {
    *ReadFlag = 0;      move16();
    *AppendFlag = 0;    move16();
    test();
    if(strchr(FormatString, (Word16)'a') != NULL)
    {
      *AppendFlag = 1;    move16();
    }
    test();
    if(strchr(FormatString, (Word16)'A') != NULL)
    {
      *AppendFlag = 1;    move16();
    }
    test();
    if(strchr(FormatString, (Word16)'r') != NULL)
    {
      *ReadFlag = 1;    move16();
    }
    test();
    if(strchr(FormatString, (Word16)'R') != NULL)
    {
      *ReadFlag = 1;    move16();
    }
  }
}

