#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#include "amr_plus_fx.h"


void Init_decoder_amrwb_plus(
  Decoder_State_Plus_fx *st,      /* (i/o) :  decoder memory state      */
  Word16 num_chan,                /* (i)   :  Number of channel         */
  Word16 fscale,                  /* (i)   ;  Frequency scaling         */
  Word16 full_reset               /* (i)   : Do entire reset au decoder */
)
{
  
  /* initialize memories (stereo part) */
  
  st->left.wover_frac = 0;    move16();
  st->right.wover_frac = 0;   move16();
  test();
  if (fscale == 0) 
  {
    st->left.Mean_isf_hf = Mean_isf_hf_low_rate;   move16();
    st->left.Dico1_isf_hf =Dico1_isf_hf_low_rate;  move16();
    st->right.Mean_isf_hf = Mean_isf_hf_low_rate;  move16();
    st->right.Dico1_isf_hf =Dico1_isf_hf_low_rate;  move16();
  }
  else 
  { 
    st->left.Mean_isf_hf = Mean_isf_hf_12k8;  move16();
    st->left.Dico1_isf_hf =Dico1_isf_hf_12k8; move16();
    st->right.Mean_isf_hf = Mean_isf_hf_12k8; move16();
    st->right.Dico1_isf_hf =Dico1_isf_hf_12k8; move16();
  }
  test();
  if (full_reset>0) 
  {
    Init_bass_postfilter(st);
    /* init filters memories */  
    Set_zero(st->left.wmem_oversamp, L_MEM_JOIN_OVER);
    Set_zero(st->right.wmem_oversamp, L_MEM_JOIN_OVER);
    Set_zero(st->left.wmem_oversamp_hf, 2*L_FILT);
    Set_zero(st->right.wmem_oversamp_hf, 2*L_FILT);

    Set_zero(st->left.wold_synth_hf, (D_BPF +L_SUBFR+ L_BSP + 2*D_NC + L_FDEL + 32*D_STEREO_TCX/5));
    Set_zero(st->right.wold_synth_hf, (D_BPF +L_SUBFR+ L_BSP + 2*D_NC + L_FDEL + 32*D_STEREO_TCX/5));

    Init_decoder_hf(&(st->left));
    Init_decoder_hf(&(st->right));

    Set_zero(st->wmem_sig_out,6);
    Set_zero(st->left.wmem_sig_out,6);
    Set_zero(st->right.wmem_sig_out,6);

    /* initialize memories (mono part) */
    st->last_mode = 0;      move16();       /* set last mode to acelp */
    st->wmem_deemph = 0;    move16();

    Set_zero(st->h_fx,ECU_WIEN_ORD+1);

    Init_decoder_lf(st);
    test();
    if(sub(num_chan,2) == 0)
    {
      Init_decoder_stereo_x(st);
    }
    Set_zero(st->wold_xri,L_TCX);	

    Set_zero(st->wmem_gain_code, 4);
    Set_zero(st->wmem_lpc_hf + 1, MHF);
    st->wmem_lpc_hf[0] = 4096;     move16();
    st->wmem_gain_hf = 0;       move16();
    st->wramp_state = 64;       move16();
    st->L_gc_thres = 0;         move16();
    st->Old_bpf_scale = 0;      move16();

  }
  return;
}



/*-----------------------------------------------------------------*
 *                  Principal decoder routine                      *
 *-----------------------------------------------------------------*/

Word16 Decoder_amrwb_plus(    /* (o)  : number of sample processed */
  Word16 codec_mode,          /* (i)  : AMR-WB+ mode (see cnst.h)  */
  Word16 serial[],            /* (i)  : serial parameters (4x20ms) */
  Word16 bad_frame[],         /* (i)  : bfi (bad_frame[4])         */
  Word16 L_frame,             /* (i)  : frame size of synthesis    */
  Word16 n_channel,           /* (i)  : 1 or 2 (mono/stereo)       */
  Word16 channel_right[],     /* (o)  : used on mono and stereo    */
  Word16 channel_left[],      /* (o)  : used on stereo only        */
  Decoder_State_Plus_fx *st,  /* (i/o): decoder memory state       */
  Word16 fscale,              /* (i)  : Frequency scaling          */
  Word16 StbrMode,            /* (i)  : Stereo bitrate mode        */   
  Word16 mono_dec_stereo,     /* (i)  : 1=Mono decoding with stereo bitstream */
  Word16 upsamp_fscale        /* (i)  : Upsampling frequency scaling */
)
{
  Word16 AqLF[(NB_SUBFR+1)*(M+1)];     /* LPC coefficients for band 0..6400Hz */
  Word16 sig_left[L_FRAME_PLUS+2*L_FDEL+20]; /* Needed to extend sig_left in order to use it in decoder_stereo_x*/
  Word16 synth_buf[L_BSP+L_TCX];
  
  Word16 pit_gain_[NB_SUBFR];
  
  Word16 pitch[NB_SUBFR];
  Word16 nbits_AVQ[4];      
  Word16 param[DEC_NPRM_DIV*NB_DIV];
  Word16 prm_stereo[MAX_NPRM_STEREO_DIV*NB_DIV];   /* see cnst.h */
  Word16 prm_hf_left[NPRM_BWE_DIV*NB_DIV];
  Word16 prm_hf_right[NPRM_BWE_DIV*NB_DIV];

  /* Scalars */
  Word16 i, k;
  Word16 mod_buf[1+NB_DIV], *mod;
  Word16 nbits_pack;
  Word16 bad_frame_hf[4];
  Word16 n_loss, any_tcx80;
  Word16 nb_samp;

  Word16 *synth;

  synth = synth_buf + L_FILT_2k;


  /* for 20-ms packetization, divide by 4 the 80-ms bitstream */
  nbits_pack = shr(add(NBITS_CORE_FX[codec_mode],NBITS_BWE),2);  

  mono_dec_stereo = 0;        move16();
  test();
  if (StbrMode >= 0) 
  {
    test();
    if (sub(n_channel,1) == 0)
    {
      mono_dec_stereo = 1;    move16();
    }
  
    nbits_pack = add(nbits_pack, shr(add(StereoNbits_FX[StbrMode], NBITS_BWE), 2));  
  }

  /*------------------------------------------------------------------------*
   * read modes (somes modes may be unreadable if the related frame is lost)*
   *------------------------------------------------------------------------*
   * mode 0 = ACELP 20ms   --> mode[] = 0,x,x,x  ..  x,x,x,0                *
   * mode 1 = TCX 20ms     --> mode[] = 1,x,x,x  ..  x,x,x,1                *
   * mode 2 = TCX 40ms     --> mode[] = 2,2,x,x  or  x,x,2,2                *
   * mode 3 = TCX 80ms     --> mode[] = 3,3,3,3                             *
   *------------------------------------------------------------------------*/

  mod = mod_buf+1;
  mod[-1] = st->last_mode;    move16(); /* previous mode */
  
  n_loss = 0;                 move16();	

  for (k=0; k<NB_DIV; k++) 
  {
    test();  
    if (bad_frame[k] == 0) 
    {
      mod[k] = Bin2int(2, &serial[k*nbits_pack]); 
    }
    else
    {
      mod[k] = -1;
    }
    n_loss = add(n_loss,bad_frame[k]);
  }

  for(i=0;i<4;i++)
  {
    bad_frame_hf[i] = bad_frame[i];   move16();
  }
  any_tcx80 = (mod[0] == 3);      logic16(); 

  for (i=1; i<4; i++) 
  {
    any_tcx80 |= (mod[i]==3);     logic16(); logic16(); 
  }
  test();test();
  if ((sub(n_loss,2) > 0) && (any_tcx80)) 
  {
    for (i=0; i<4; i++) 
    {
      mod[i] = -1;                move16();
      bad_frame[i] = 1;           move16();   
    }
  }

  /* extrapolate mode in case of packet erasures and fix bit errors */
  test();test();test();test();
  if ( (sub(mod[0],3)==0) || (sub(mod[1],3)==0) || (sub(mod[2],3)==0) || (sub(mod[3],3)==0)) 
  {
    /* handle loss of one or several TCX-80 packets */ 
    for (k=0; k<NB_DIV; k++) 
    {
      mod[k] = 3;                 move16();
    }
  }
  else
  {
    /* handle loss of a TCX-40 packet */
    test();test();
    if ((sub(mod[0],2) == 0) || (sub(mod[1],2) == 0)) 
    {
      for (k=0; k<2; k++) 
      {
        mod[k] = 2;               move16();
      }
    }
    test();test();
    if ((sub(mod[2],2) == 0) || (sub(mod[3],2) == 0)) 
    {
      for (k=2; k<4; k++) 
      {
        mod[k] = 2;               move16();
      }
    }

    /* handle loss of an ACELP or TCX-20 packet */

    for (k=0; k<NB_DIV; k++) 
    {
      test();
      if (mod[k] < 0) 
      {
        /* repeat the previous mode:
           if previous mode = ACELP        -> ACELP
           if previous mode = TCX-20/40/80 -> TCX-20
           notes:
           - ACELP is not allowed after TCX (no pitch information to reuse)
           - TCX-40 is not allowed in the mode repetition to keep the logic simple */
        test();
        if (mod[k-1] == 0) 
        {
          mod[k] = 0;               move16();
        }
        else
        {
          mod[k] = 1;               move16();
        }
      }
    }
  }

  /*----------- DECODE BIT-STREAM ----------*/

  Dec_prm(mod, bad_frame, serial, nbits_pack, codec_mode, param, nbits_AVQ);                    

  test();test();                                                                                              
  if ((sub(n_channel,2) == 0) && (StbrMode >= 0))
  {
    Dec_prm_stereo_x(bad_frame_hf, serial, nbits_pack, NBITS_BWE, prm_stereo, StbrMode, st);    
  }
  Dec_prm_hf(mod, bad_frame, serial, nbits_pack, prm_hf_right);                                 

  test();                                                                                              
  if (sub(n_channel,2) == 0 || sub(mono_dec_stereo,1) ==0) 
  {
    test();
    if (StbrMode < 0)
    {
      Copy(prm_hf_right, prm_hf_left, NPRM_BWE_DIV*NB_DIV);
    }
    else
    {
      Dec_prm_hf(mod, bad_frame, serial-NBITS_BWE/4, nbits_pack, prm_hf_left);
    }
  }

  Decoder_amrwb_plus_1(channel_right,channel_left,mod,param,prm_hf_right,prm_hf_left,nbits_AVQ,
       codec_mode,bad_frame,bad_frame_hf,AqLF,synth,pitch,pit_gain_,st,n_channel,
       L_frame, fscale,mono_dec_stereo);

  test();
  if (sub(n_channel,2) == 0)
  {
    Decoder_stereo_x(prm_stereo, bad_frame_hf, sig_left, synth, AqLF,StbrMode,fscale, st);  
    test();
    if (fscale == 0)
    {
      HP50_12k8(sig_left, L_FRAME_PLUS, st->right.wmem_sig_out);                             
      HP50_12k8(synth, L_FRAME_PLUS, st->left.wmem_sig_out);                                 

      Oversamp_12k8(sig_left, channel_left, L_frame, st->left.wmem_oversamp, 0,1);              
      Oversamp_12k8(synth, channel_right, L_frame, st->right.wmem_oversamp, 0,1);               
      nb_samp = L_frame;    move16();

    }
    else 
    {
      /* band join and oversampling (HF into channel_left/right) */
      /*fac_fs = (L_FRAME48k*upsamp_fscale)/L_frame;*/

      nb_samp = Join_over_12k8(sig_left, channel_left, L_FRAME_PLUS, channel_left,L_frame,
                     upsamp_fscale, st->left.wmem_oversamp, &(st->left.wover_frac));
      nb_samp = Join_over_12k8(synth, channel_right, L_FRAME_PLUS, channel_right,L_frame,
                     upsamp_fscale, st->right.wmem_oversamp, &(st->right.wover_frac));
    }
  }
  else
  {
   test();
    if (fscale == 0)
    {
      Oversamp_12k8(synth, channel_right, L_frame, st->right.wmem_oversamp, 0,1);               
      nb_samp = L_frame;      move16();
    }
    else 
    {
      /* band join and oversampling (HF into channel_left/right) */
      /*fac_fs = (L_FRAME48k*upsamp_fscale)/L_frame;*/

      nb_samp = Join_over_12k8(synth, channel_right, L_FRAME_PLUS, channel_right,L_frame,
                     upsamp_fscale, st->right.wmem_oversamp, &(st->right.wover_frac));
    }

  }

  /* update for next superframe */
  st->last_mode = mod[NB_DIV-1];  move16();

  return(nb_samp);
}


void Decoder_amrwb_plus_1(
    Word16* chan_right,     /* (o)  : Channel right synthesis HF Q_syn +1 */
    Word16* chan_left,      /* (o)  : Channel left synthesis  HF Q_syn +1 */  
    Word16* mod,            /* (i)  : mode for each 20ms frame (mode[4])  */
    Word16* param,          /* (i)  : parameter to decode                 */    
    Word16* prm_hf_right,   /* (i)  : HF parameters channel right         */
    Word16* prm_hf_left,    /* (i)  : HF parameters channel left          */    
    Word16* nbits_AVQ,      /* (i)  : Number of bits used by AVQ_dec      */  
    Word16  codec_mode,     /* (i)  : AMR-WB+ mode (see cnst.h)           */ 
    Word16* bad_frame,      /* (i)  : for each frame (bad_frame[4])       */
    Word16* bad_frame_hf,   /* (i)  : for each frame (bad_frame[4])       */  
    Word16* AqLF,           /* (o)  : decoded coefficients (AdLF[16])     */  
    Word16* synth,          /* (o)  : decoded synthesis              Q_syn*/
    Word16* pitch,          /* (o)  : decoded pitch (pitch[16])           */
    Word16* pit_gain,       /* (o)  : decoded pitch gain (pit_gain[16])Q14*/
    Decoder_State_Plus_fx* st, /* (i/o) : coder memory state              */
    Word16 n_channel,       /* (i)  : Number of channels                  */
    Word16 L_frame,         /* (i)  : Frame length                        */
    Word16 fscale,          /* (i)  : frequency scaling                   */    
    Word16 mono_dec_stereo  /* (i)  : 1=Mono decoding with stereo bitstream */
)
{

  Word16 exc_buf[PIT_MAX_MAX+L_INTERPOL+L_TCX]; 
  Word16 *exc;

  exc = exc_buf + PIT_MAX_MAX + L_INTERPOL;
  Copy(st->wold_exc, exc_buf, PIT_MAX_MAX + L_INTERPOL);

  Decoder_lf(mod, param, nbits_AVQ, codec_mode, bad_frame, AqLF, exc, synth, pitch, pit_gain,fscale, st );		

  Scale_mem2(synth, st->old_syn_q, st, 1);
  Copy(&synth[L_FRAME_PLUS-M], st->cp_old_synth, M);        /* needed in cp state */



  Copy(exc_buf+L_FRAME_PLUS, st->wold_exc, PIT_MAX_MAX+L_INTERPOL);

  exc = exc_buf;

  Scale_mem2(exc, st->old_subfr_q, st, 0);
  Deemph(synth, PREEMPH_FAC_FX, L_FRAME_PLUS, &(st->wmem_deemph));
  Bass_postfilter(synth, pitch, pit_gain, synth, fscale, st);

  test();
  if (fscale == 0)
  {
    test();
    if (sub(n_channel,1) == 0) 
    {
      /* high pass filter only in the mono case to avoid phase mismatch */
      HP50_12k8(synth, L_FRAME_PLUS, st->wmem_sig_out);
    }
  }
 
  test();  
  if (sub(L_frame,L_FRAME8k) >0) 
  {
  
    Decoder_hf(mod, prm_hf_right, prm_hf_left, mono_dec_stereo, bad_frame_hf, AqLF, exc, chan_right, st->wmem_lpc_hf, &(st->wmem_gain_hf), &(st->wramp_state), &(st->right),  st->Old_Q_exc);
    test();
    if (sub(n_channel,1) == 0) 
    {
      test();
      if(fscale == 0) 
      {
        Delay(chan_right, L_FRAME_PLUS, DELAY_PF, st->right.wold_synth_hf);
      }
      else 
      {
        Delay(chan_right, L_FRAME_PLUS, DELAY_PF+L_SUBFR, st->right.wold_synth_hf);
      }
    }
    else  
    {
      test();
      if(fscale == 0) 
      {
        Delay(chan_right, L_FRAME_PLUS, (D_BPF+L_BSP+2*D_NC+L_FDEL+32*D_STEREO_TCX/5) , st->right.wold_synth_hf);	
      }
      else 
      {
        Delay(chan_right, L_FRAME_PLUS, (D_BPF+L_SUBFR+L_BSP+2*D_NC+L_FDEL+32*D_STEREO_TCX/5) , st->right.wold_synth_hf);		
      }
    }
    test();
    if (sub(n_channel,2) == 0) 
    {
      Decoder_hf(mod, prm_hf_left, prm_hf_right, mono_dec_stereo, bad_frame_hf, AqLF, exc, chan_left, st->wmem_lpc_hf, &(st->wmem_gain_hf), &(st->wramp_state), &(st->left),  st->Old_Q_exc);
      test();
      if(fscale == 0) 
      {
        Delay(chan_left, L_FRAME_PLUS,(D_BPF + L_BSP + 2*D_NC + L_FDEL + 32*D_STEREO_TCX/5), st->left.wold_synth_hf);
      }
      else 
      {
        Delay(chan_left, L_FRAME_PLUS,(D_BPF + L_SUBFR + L_BSP + 2*D_NC + L_FDEL + 32*D_STEREO_TCX/5), st->left.wold_synth_hf);
      }
    }

    test();
    if (fscale == 0)
    {
      Oversamp_12k8(chan_right, chan_right, L_frame, st->right.wmem_oversamp_hf, 1,0);    
      test();
      if (sub(n_channel,2) == 0) 
      {
        Oversamp_12k8(chan_left, chan_left, L_frame, st->left.wmem_oversamp_hf, 1,0);     
      }
    }
  }
  else
  {
    Set_zero( chan_right, L_frame );                                                          
    Set_zero( chan_left, L_frame );                                                           
  }

  return;
}

